// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// Bomber & Bomb.

struct OBomb : TPlayObject {
  public:
  int x;
  int y;
  int xtime;
  int range;

  OBomb(int _owner, int _x, int _y, int _xtime, int _range) : x(_x), y(_y) {
    owner = _owner;
    range = _range;
    time = g.time;
    xtime = g.time + _xtime;
    g.mapSet(x,y, cBomb1, owner);
    g.objects.push_back(this);
    }
  OBomb(TBuffer &b) {
    b.get(owner);
    b.get(x); b.get(y); b.get(time); b.get(xtime); b.get(range);
    }
  void save(TBuffer &b) {
    b.putChar(11);
    b.put(owner); b.put(x); b.put(y); b.put(time); b.put(xtime); b.put(range);
    }
  void go() {
    if(time < xtime && g.map[y][x] != cExplosion) {
      int phase = (2500 / (xtime-time));
      g.mapSet(x, y, phase&1 ? cBomb1 : cBomb2, owner);
      time += 25;
      if(time > xtime) time = xtime;
      }
    else if(time <= xtime) {
      g.mapSet(x,y, cExplosion, owner);
      for(int k=0; k<4; k++)
      for(int r=1; r<=range; r++) {
        int ax = x+kx[k]*r, ay = y+ky[k]*r;
        CellType m = bigType(ax, ay);
        if(isBonus(m)) {
          clearBig(ax, ay, cFree, 0);
          }
        if(isCannon(m)) {
          clearBig(ax, ay, cFree, 0);
          g.mapSet(ax,ay, cExplosion, owner);
          r=range;
          }
        else switch(g.map[ay][ax]) {
          case cExplosion:
            // might be someone else's
            break;
          case cFree:
          case cBomb1:
          case cBomb2:
            g.mapSet(ax,ay, cExplosion, owner);
            break;
          case cBreakWall:
          case cSnakeTRON:
          case cTetrisDown:
          case cTetris:
          case cCastleWall:
            g.mapSet(ax,ay, cExplosion, owner);
            r=range;
            break;
          case cSnakeHead:
            g.mapSet(ax,ay, cSnakeDead, owner);
            r=range;
            break;
          default:
            r=range;
            break;
          }
        }
      xtime = time; // if it exploded due to another's explosion
      time += 750;
      ow()->ammo += 10;
      }
    else if(time > xtime) {
      g.mapSet(x,y, cFree, 0);
      for(int k=0; k<4; k++)
      for(int r=1; r<=range; r++) {
        int ax = x+kx[k]*r, ay = y+ky[k]*r;
        if(ax == 0 || ay == 0 || ax == g.sizeX-1 || ay == g.sizeY-1) continue;
        if(g.map[ay][ax] == cExplosion) 
          g.mapSet(ax,ay, cFree, 0);
        }
      g.deleteObject(this);
      }
    }
  };

struct OBomber : TPlayObject {
  public:
  int x;
  int y;
  OBomber(int _owner, int _x, int _y) : 
    x(_x), y(_y) {
    owner = _owner;
    time = g.time;
    g.objects.push_back(this);
    }
  
  OBomber(TBuffer &b) {
    b.get(owner); 
    b.get(x); b.get(y);
    b.get(time);
    }
  void save(TBuffer &b) {
    b.putChar(10);
    b.put(owner); 
    b.put(x); b.put(y);
    b.put(time);
    }

  bool checkGo(int cx, int cy) {
    cx = DeCtr(cx);
    cy = DeCtr(cy);
    // cannot block self with bomb
    if(cx == DeCtr(x) && cy == DeCtr(y)) return true;
    if(g.map[cy][cx] == cFree) return true;
    if(g.map[cy][cx] == cExplosion) return true;
    CellType m = bigType(cx, cy);
    if(isBonus(m)) return true;
    return false;
    }
  
  void tryToGo(int k) {
    int rx = x + kx[k] * (384 + 128);
    int ry = y + ky[k] * (384 + 128);
    bool u0 = checkGo(rx - ky[k] * 384, ry + kx[k] * 384);
    bool u1 = checkGo(rx + ky[k] * 384, ry - kx[k] * 384);
    if(u0 && u1) {
      x += kx[k] * 128;
      y += ky[k] * 128;
      }
    }
  
  void go() {
    for(int k=0; k<4; k++) if(g.pl[owner]->key[k]) tryToGo(k);
    time += spd()/2;
    int nx = DeCtr(x);
    int ny = DeCtr(y);
    CellType m = bigType(nx, ny);
    if(isBonus(m)) {
      switch(m) {
        case cBonusAmmo:
          ow()->score += scoreGain();
          ow()->ammo += 10;
          break;
        case cBonusLevel:
          ow()->score += scoreGain() * 10;
          ow()->level += 10;
          break;
        case cBonusScore:
          ow()->score += scoreGain();
          ow()->typeval ++;
          break;
        }
      placeBigRandomly(m, 0);
      clearBig(nx, ny, cFree, 0);
      }
    if(g.map[ny][nx] != cFree && g.map[ny][nx] != cBomb1 && g.map[ny][nx] != cBomb2) {
      killedBy(g.owner[ny][nx]);
      return;
      }
    }

  void draw() {
    d.setIconColor(ow()->color);
    d.bomber.drawAtCentered(CTG(x), CTG(y));
    }
  void reactToKey(PlayerKey pk, bool pressed) {
    if(!pressed) return;
    int nx = DeCtr(x);
    int ny = DeCtr(y);
    if(pk == keyFire && g.map[ny][nx] == cFree && ow()->useAmmo(10))
      new OBomb(owner, nx, ny, 50 * spd(), ow()->typeval);
    }
  };
