// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// Interface (both game and menu and general).

CellType lm[MAXY][MAXX];
int lo[MAXY][MAXX];

void bufferTest();

int randColor() {
  return (rand() % 256) + ((rand() % 256) << 8) + ((rand() % 256) << 16);
  }

char *qitoa(int i) {
  static char nbuf[16];
  sprintf(nbuf, "%d", i);
  return nbuf;
  }

void TDisplay::setIconColor(int color) {
  pixelcolor['P'] = color;
  pixelcolor['p'] = (color & 0xFEFEFE) >> 1;
  }

void TDisplay::init() {
  if (SDL_Init(SDL_INIT_VIDEO) < 0) {
    printf("Failed to initialize SDL.\n");
    SDL_Quit();
    }

  SDL_Quit();

  sizeX = c.resX;
  sizeY = c.resY;
  stop  = false;
  
  surface = 
    SDL_SetVideoMode(sizeX, sizeY, 32, c.fs ? SDL_FULLSCREEN : 0);

  if(!surface) {
    sizeX = 800;
    sizeY = 600;
    surface = SDL_SetVideoMode(800, 600, 32, 0);
    }
    
  if(!surface) {
    printf("Failed to initialize SDL.\n");
    SDL_Quit();
    }
  SDL_WM_SetCaption("Untahris", 0);

  if(TTF_Init() != 0) {
    printf("Failed to initialize TTF.\n");
    exit(2);
    }
  font = TTF_OpenFont("verabd.ttf", 16);
  }

void TDisplay::loop() {
  SDL_LockSurface(surface);
  for(int y=0; y<sizeY; y++)
  for(int x=0; x<sizeX; x++) {
    int u = (x+2*y) ^ (y-2*x);
    putpix(x, y, 0x101 * (u & 0x3F));
    }
  SDL_UnlockSurface(surface);
  for(int y=0; y<g.sizeY; y++)
  for(int x=0; x<g.sizeX; x++) lm[y][x] = CellType(-1);
  
  while(!stop) {
    SDL_Event event;
    bool ev = SDL_PollEvent(&event);
    if(ev) switch (event.type) {

      case SDL_KEYDOWN:
        addKeypress(event.key.keysym.sym, true);
        break;

      case SDL_KEYUP:
        addKeypress(event.key.keysym.sym, false);
        break;

      case SDL_QUIT:
        stop = true;
        return;
      }
    else {
      draw();
      dataTransfer();
      int t = getCurrentTime();
      g.upTime(t);
      SDL_Delay(1);
      }
    }
  }

void TDisplay::addKeypress(int key, bool type) {
  for(int i=0; i<size(lps); i++) {
    TLocalPlayer& l(lps[i]);
    for(int u=0; u<8; u++) if(l.keys[u] == key) {
      TKeypress k;
      k.time = getCurrentTime()+1;
      k.pressed = type;
      k.player = l.index;
      k.key = PlayerKey(u);
      addKeyToQueue(k);
      broadcastKey(k);
      }
    }
  if(key == 27) stop = true;
  if(key == 'y') bufferTest();
  }

void TDisplay::putpix(int x, int y, int c) {
  if(x<0 || y<0 || x>=sizeX || y>=sizeY) return;
  
  if(x >= ofsX && y >= ofsY) {
    int zx = (x-ofsX) / d.Res;
    int zy = (y-ofsY) / d.Res;
    if(zx<g.sizeX && zy<g.sizeY) lm[zy][zx] = CellType(-1);
    }
  
  unsigned char *dst = (unsigned char*) surface->pixels;
  dst += y*surface->pitch;
  int* ptr = (int*) dst;
  ptr[x] = c;
  }

void TDisplay::draw() {
  SDL_LockSurface(surface);

  for(int y=0; y<g.sizeY; y++)
  for(int x=0; x<g.sizeX; x++) if(g.map[y][x] != lm[y][x] || g.owner[y][x] != lo[y][x]) {
    int tx = ofsX + x * Res;
    int ty = ofsY + y * Res;
    CellType m = g.map[y][x];
    setIconColor(g.pl[g.owner[y][x]]->color);
    if(m == cBig) {
      for(int ky=0; ky<2; ky++) for(int kx=0; kx<2; kx++) {
        m = g.map[y-ky][x-kx];
        if(isBonus(m) || isCannon(m)) {
          tx -= Res * kx;
          ty -= Res * ky;
          goto bre;
          }
        }
      bre: ;
      }
    if(m == cFree) freecell.drawAt(tx, ty);
    else if(m == cBonusLevel) bonusLevel.drawAt(tx, ty);
    else if(m == cBonusAmmo) bonusAmmo.drawAt(tx, ty);
    else if(m == cBonusScore) bonusScore.drawAt(tx, ty);
    else if(m == cCannonF) cannon.drawAt(tx, ty);
    else if(m == cCannonA) cannona.drawAt(tx, ty);
    else if(m == cPermaWall) permawall.drawAt(tx, ty);
    else if(m == cBomb1) bomb1.drawAt(tx, ty);
    else if(m == cBomb2) bomb2.drawAt(tx, ty);
    else if(m == cExplosion) explosion.drawAt(tx, ty);
    else breakwall.drawAt(tx, ty);

    lo[y][x] = g.owner[y][x];
    lm[y][x] = g.map[y][x];
    }
  
  for(int o=0; o<size(g.objects); o++)
    if(g.objects[o]) g.objects[o]->draw();

  // do not draw dummy
  for(int p=1; p<size(g.pl); p++) g.pl[p]->draw();

  SDL_UnlockSurface(surface);

  int y0 = sizeY - 20 * size(g.pl) - 10;
  int step = (sizeX-20)/7;
  putText(10, y0, 0xFFFFFF, "TYPE", 0);
  putText(10+3*step, y0, 0xFFFFFF, "LIVES", 2);
  putText(10+4*step, y0, 0xFFFFFF, "FRAGS", 2);
  putText(10+5*step, y0, 0xFFFFFF, "SCORE", 2);
  putText(10+6*step, y0, 0xFFFFFF, "LEVEL", 2);
  putText(10+7*step, y0, 0xFFFFFF, "AMMO", 2);
  for(int p=1; p<size(g.pl); p++) {
    TPlayer *pla = g.pl[p];
    int c = pla->color;
    int y = y0 + 20*p;

    for(int u=2; u<8; u++) {
      putText(10+u*step, y, 0, "||||||||||||||", 2);
      putText(10+u*step+2, y, 0, "||||||||||||||", 2);
      }
    putText(10, y, 0, "|||||||||||||||||||||||||||||||", 0);
    putText( 8, y, 0, "|||||||||||||||||||||||||||||||", 0);

    putText(10, y, c, pla->getTypeName(), 0);
    if(pla->typeval != -1) putText(10+2*step, y, c, qitoa(pla->typeval), 2);
    putText(10+3*step, y, c, qitoa(pla->lives), 2);
    putText(10+4*step, y, c, qitoa(pla->frags), 2);
    putText(10+5*step, y, c, qitoa(pla->score), 2);
    putText(10+6*step, y, c, qitoa(pla->level), 2);
    putText(10+7*step, y, c, qitoa(pla->ammo), 2);
    }

  SDL_UpdateRect(surface, 0, 0, 0, 0);
  }

void TDisplay::putText(int x, int y, int color, const char *text, int align) {
  if(text[0] == 0) return;
  if(d.font) {
    SDL_Color col;
    col.r = 0;
    col.g = 0;
    col.b = 0;
    SDL_Surface *txt = TTF_RenderText_Solid(d.font, text, col);
    SDL_Rect rect;
    rect.y = y;
    rect.h = sizeY-y;
    rect.w = txt->w;
    switch(align) {
      case 0: rect.x = x; break;
      case 1: rect.x = x - rect.w/2; break;
      case 2: rect.x = x - rect.w; break;
      }
      
    rect.x--; SDL_BlitSurface(txt, NULL,d.surface,&rect); rect.x++;
    rect.x++; SDL_BlitSurface(txt, NULL,d.surface,&rect); rect.x--;
    rect.y--; SDL_BlitSurface(txt, NULL,d.surface,&rect); rect.y++;
    rect.y++; SDL_BlitSurface(txt, NULL,d.surface,&rect); rect.y--;
    SDL_FreeSurface(txt);
    
    col.r = color >> 16;
    col.g = color >> 8;
    col.b = color;
    txt = TTF_RenderText_Solid(d.font, text, col);
    SDL_BlitSurface(txt, NULL,d.surface,&rect);
    SDL_FreeSurface(txt);
    }
  }

void TDisplay::menuLoop() {
  while(!stop) {
    SDL_Event event;
    bool ev = SDL_PollEvent(&event);
    if(ev) switch (event.type) {

      case SDL_KEYDOWN:
        menuKey(event.key.keysym.sym);
        break;

      case SDL_QUIT:
        stop = true;
        return;
      }
    else
      menuDraw();
    }
  }

int goodPos(int i) { return d.menupos == i ? 0xFFFFFF : 0x0000FF; }

char *gameTypeName(TGameType t) {
  if(t == gLocal) return "LOCAL";
  if(t == gHost) return "SERVER";
  if(t == gClient) return "CLIENT";
  return "UNKNOWN";
  }

void drawMenuLine(int y,const char *text, const char *option) {
  int r = d.sizeX / 2;
  d.putText(r-20, 80+20*y, goodPos(y), text, 2);
  d.putText(r+20, 80+20*y, goodPos(y), option, 0);
  }

void drawMenuLine(int y,const char *text, int val) {
  char buf[16];
  sprintf(buf, "%d", val);
  drawMenuLine(y, text, buf);
  }

void TDisplay::menuDraw() {
  SDL_LockSurface(surface);

  for(int y=0; y<sizeY; y++)
  for(int x=0; x<sizeX; x++) putpix(x, y, (((x ^ y) & 0x3F) << 8));

  SDL_UnlockSurface(surface);

  int r = sizeX / 2;
  putText(r, 20, 0x00FF00, "UNTAHRIS", 1);
  putText(r, 40, 0x00FF00, "Common Playground", 1);
  
  drawMenuLine(0, "Multiplayer (L/S/C)", gameTypeName(c.gt));
  drawMenuLine(1, "Server", c.server.c_str());
  drawMenuLine(2, "port", c.port);
  drawMenuLine(3, "screen resolution X", c.resX);
  drawMenuLine(4, "screen resolution Y", c.resY);
  drawMenuLine(5, "cells X", c.sizeX);
  drawMenuLine(6, "cells Y", c.sizeY);
  drawMenuLine(7, "cell resolution", c.res);
  drawMenuLine(8, "full screen", c.fs ? "YES" : "NO");
  drawMenuLine(9, "centered", c.centered ? "YES" : "NO");
  for(int p=0; p<size(lps); p++) {
    bool onP = menupos == 10+p;
    char buf[80];
    if(!onP) sprintf(buf, "edit a player");
    else sprintf(buf, "Change color (RGB), redefine Keys, Delete");
    putText(r, 280 + 20*p, lps[p].color, buf, 1);
    }
  putText(r, 280 + 20*size(lps), goodPos(10+size(lps)), "create a new player", 1);
  
  putText(r, d.sizeY-40, 0xFF0000, c.err.c_str(), 1);

  SDL_UpdateRect(surface, 0, 0, 0, 0);
  }

void checkConfig() {
  c.err = "Status OK";
  c.ok = true;
  if(c.sizeX * c.res > d.sizeX || c.sizeY * c.res > d.sizeY)
    c.ok = false, c.err = "Would not fit";
  else if(c.sizeX < 10 || c.sizeY < 10)
    c.ok = false, c.err = "Screen too small";
  else if(c.sizeX > MAXX || c.sizeY > MAXY)
    c.ok = false, c.err = "Screen too big";
  else if(c.res < 1)
    c.ok = false, c.err = "Bad cell resolution";
  else if(c.res > 32)
    c.ok = false, c.err = "Bad cell resolution (32 is max)";
  }

void affectString(int key, string& s) {
  if(key > 128) return;
  if(key == 8) {
    if(s.size() <= 1) s = "";
    else s.erase(s.size() - 1);
    }
  else s += char(key);
  }

template<class T> void affectToggle(int key, T& x, int vals) {
  char* ptr = (char*) (&x);
  (*ptr)++;
  (*ptr) %= vals;
  }

void affectGameType(int key, TGameType& gt) {
  if(key == 'l') gt = gLocal;
  if(key == 'h') gt = gHost;
  if(key == 's') gt = gHost;
  if(key == 'c') gt = gClient;
  if(key == 'j') gt = gClient;
  }

void affectInt(int key, int& val) {
  if(key >= '0' && key <= '9') {
    val *= 10;
    val += key - '0';
    }
  else if(key == 8) val /= 10;
  else if(key == '+' || key == SDLK_KP6 || key == SDLK_RIGHT)
    val++;
  else if(key == '-' || key == SDLK_KP4 || key == SDLK_LEFT)
    val--;
  }

void editKeysFor(TLocalPlayer& lp) {
  char *keyNames[8] = {
    "UP",
    "RIGHT",
    "DOWN",
    "LEFT",
    "FIRE",
    "EARTH",
    "AIR",
    "WATER"
    };
  for(int k=0; k<8; k++) {
    SDL_LockSurface(d.surface);
  
    for(int y=0; y<d.sizeY; y++)
    for(int x=0; x<d.sizeX; x++) d.putpix(x, y, (((x+y)^(x-y))&0xFF) << 16);
  
    SDL_UnlockSurface(d.surface);
  
    d.putText(d.sizeX / 2, d.sizeY / 2 - 10, 0x00FFFF, keyNames[k], 1);

    SDL_UpdateRect(d.surface, 0, 0, 0, 0);

    while(true) {
      SDL_Event event;
      bool ev = SDL_PollEvent(&event);
      if(ev) switch (event.type) {
  
        case SDL_KEYDOWN:
          lp.keys[k] = event.key.keysym.sym;
          if(lp.keys[k] == 27) return;
          goto nextKey;
  
        case SDL_QUIT:
          return;
        }
      }
    nextKey: ;
    }
  }

void affectPlayer(int key, int index) {
  switch(key) {
    case 'c':
      lps[index].color = randColor();
      break;
    case 'r':
      lps[index].color ^= (rand() & 0xFF) << 16;
      break;
    case 'g':
      lps[index].color ^= (rand() & 0xFF) << 8;
      break;
    case 'b':
      lps[index].color ^= (rand() & 0xFF);
      break;
    case 'k':
      editKeysFor(lps[index]);
      break;
    case 8:
    case 'd':
    case 'e':
      lps.erase(lps.begin()+index, lps.begin()+index+1);
      break;
    }
  }

void TDisplay::menuKey(int key) {
  switch(key) {
    case 13: {
      startGame();
      stop = false;
      break;
      }
    case 27:
      stop = true;
      break;
    case SDLK_KP2: case SDLK_UP:
      menupos--;
      if(menupos < 0) menupos = 0;
      break;
    case SDLK_KP8: case SDLK_DOWN:
      menupos++;
      if(menupos > 10 + size(lps)) menupos = 10 + size(lps);
      break;
    default:
      if(menupos==0) affectGameType(key, c.gt);
      if(menupos==1) affectString(key, c.server);
      if(menupos==2) affectInt(key, c.port);
      if(menupos==3) affectInt(key, c.resX);
      if(menupos==4) affectInt(key, c.resY);
      if(menupos==5) affectInt(key, c.sizeX);
      if(menupos==6) affectInt(key, c.sizeY);
      if(menupos==7) affectInt(key, c.res);
      if(menupos==8) affectToggle(key, c.fs, 2);
      if(menupos==9) affectToggle(key, c.centered, 2);
      if(menupos>= 10 && menupos < 10 + size(lps))
        affectPlayer(key, menupos - 10);
      else if(menupos==10+size(lps)) {
        TLocalPlayer newOne;
        newOne.color = randColor();
        lps.push_back(newOne);
        editKeysFor(newOne);
        }
      if(menupos == 3 || menupos == 4 || menupos == 8)
        c.err = "Need to restart to apply screen settings";
      else
        checkConfig();
    }
  }

int TDisplay::getCurrentTime() {
  return SDL_GetTicks() + delta;
  }

void TDisplay::setCurrentTime(int val) {
  delta = val - SDL_GetTicks();
  }

void TDisplay::initForGame() {
  
  d.Res = c.res;
  if(c.centered) {
    d.ofsX = (d.sizeX - c.res * g.sizeX) / 2;
    d.ofsY = (d.sizeY - c.res * g.sizeY) / 2;
    }
  else {
    d.ofsX = 0;
    d.ofsY = 0;
    }
  
  permawall.fillWith("wPW", d.Res);
  breakwall.fillWith("pP", d.Res);
  castle.fillWith("pY0P0P0", d.Res);
  freecell.fillWith("dhph", d.Res);
  bonusAmmo.fillWith("dhhrRrpRpR", d.Res*2);
  bonusScore.fillWith("dhhgGgpGpG", d.Res*2);
  bonusLevel.fillWith("dhhbBbpBpB", d.Res*2);
  cannon.fillWith("dhhYPYPYP", d.Res*2);
  cannona.fillWith("dhhRPRPRP", d.Res*2);
  cannonp.fillWith("000RPRPRP", d.Res*2);
  spawncursor.fillWith("pPp0", d.Res+3);
  target.fillWith("PP000P", 11);
  thruster.fillWith("0", d.Res*2);
  bomb1.fillWith("dhphR", d.Res);
  bomb2.fillWith("dhphY", d.Res);
  bomber.fillWith("pP", d.Res*3/4);
  explosion.fillWith("RYRYPO", d.Res);
  bullet.fillWith("PW", 8);
  for(int y=0; y<d.Res*3/2; y++) {
    int w = (y * 14) / 30;
    for(int x=d.Res-1-w; x <= d.Res+w; x++)
      thruster.pixel[y][x] = 'P';
    }
  
  char buf[64];
  sprintf(buf, "graph%d.unt", d.Res);
  FILE *f = fopen(buf, "rt");
  if(f) {
    while(1) {
      string s = readStr(f);
      if(s == "") break;
      if(s == "PERMA") permawall.loadFrom(f);
      else if(s == "BREAK") breakwall.loadFrom(f);
      else if(s == "FREE") freecell.loadFrom(f);
      else if(s == "AMMO") bonusAmmo.loadFrom(f);
      else if(s == "LEVEL") bonusLevel.loadFrom(f);
      else if(s == "SCORE") bonusScore.loadFrom(f);
      else if(s == "THRUST") thruster.loadFrom(f);
      else if(s == "BULLET") bullet.loadFrom(f);
      else if(s == "BALL") ball.loadFrom(f);
      else if(s == "BOMBER") bomber.loadFrom(f);
      else if(s == "BOMB1") bomb1.loadFrom(f);
      else if(s == "BOMB2") bomb2.loadFrom(f);
      else if(s == "EXPLOSION") explosion.loadFrom(f);
      else if(s == "CASTLE") castle.loadFrom(f);
      else if(s == "CANNON") cannon.loadFrom(f);
      else if(s == "CANNONA") cannona.loadFrom(f);
      else if(s == "CANNONP") cannonp.loadFrom(f);
      else if(s == "TARGET") target.loadFrom(f);
      else printf("Unknown icon name '%s'\n", s.c_str());
      }
    fclose(f);
    }
  
  pixelcolor['d'] = 0;
  pixelcolor['h'] = 0x000040;
  pixelcolor['W'] = 0xFFFFFF;
  pixelcolor['w'] = 0xC0C0C0;
  pixelcolor['P'] = 0xFF0000;
  pixelcolor['p'] = 0x800000;
  pixelcolor['R'] = 0xFF0000;
  pixelcolor['r'] = 0x800000;
  pixelcolor['G'] = 0x00FF00;
  pixelcolor['g'] = 0x008000;
  pixelcolor['B'] = 0x0000FF;
  pixelcolor['b'] = 0x000080;
  pixelcolor['Y'] = 0xFFFF00;
  pixelcolor['y'] = 0x808000;
  pixelcolor['O'] = 0xFF8000;
  pixelcolor['o'] = 0x804000;
  }

void TIcon::fillWith(char *chars, int _size) {
  int len = strlen(chars);
  size = _size;
  for(int x=0; x<size; x++)
  for(int y=0; y<size; y++) {
    int k = x;
    if(y < k) k = y;
    if(size-y-1 < k) k = size-y-1;
    if(size-x-1 < k) k = size-x-1;
    if(k >= len) k = len-1;
    pixel[y][x] = chars[k];
    }
  }

void TIcon::loadFrom(FILE *f) {
  string line = readStr(f);
  if(line.size() > 64) { printf("Image too big\n"); return; }
  size = line.size();
  for(int y=0; y<size; y++) {
    if(y) line = readStr(f);
    if(line.size() > 64) { printf("Line too long\n"); return; }
    if(int(line.size()) != size) {printf("Misformed line\n"); return; }
    for(int x=0; x<int(line.size()); x++)
      pixel[y][x] = line[x];
    }
  }

void TIcon::drawAt(int ox, int oy) {
  for(int x=0; x<size; x++)
  for(int y=0; y<size; y++) {
    char ch = pixel[y][x];
    if(ch != '0')
      d.putpix(ox+x, oy+y, d.pixelcolor[ch]);
    }
  }

void TIcon::drawAtRotated(int ox, int oy, int rx, int ry) {
  for(int x=0; x<size; x++)
  for(int y=0; y<size; y++) {
    char ch = pixel[y][x];
    if(ch != '0') {
      int x0 = x*2-size;
      int y0 =  y*2-size;
      if(size&1 == 0) x0 += 1, y0 += 1;
      d.putpix(ox-((x0*ry+y0*rx+127)>>8), oy-((y0*ry-x0*rx+127)>>8), d.pixelcolor[ch]);
      }
    }
  }
  
