// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// TPlayground implementation.

void TPlayground::cleanup() {
  for(int p=0; p<size(pl); p++) delete pl[p];
  for(int o=0; o<objects.size(); o++) delete objects[o];
  objects.resize(0);
  pl.resize(0);
  }

void TPlayground::init(int sx, int sy) {
  cleanup();

  time = 0;
  kqpos = 1;
  seed = systemSeed();
  sizeX = sx;
  sizeY = sy;

  TPlayer *dummy = new TPlayer;
  addPlayer(dummy);
  dummy->color = 0xFFFFFF;
  dummy->typeval = -1;
  dummy->typenum = numTypes;

  for(int y=0; y<sy; y++) for(int x=0; x<sx; x++)
    g.mapSet(
      x, y, 
      (y == 0 || x == 0 || y == sy-1 || x == sx-1) ? cPermaWall : cFree, 
      0);
  for(int u=0; u<2; u++) {
    placeBigRandomly(cBonusScore, 0);
    placeBigRandomly(cBonusLevel, 0);
    placeBigRandomly(cBonusAmmo, 0);
    }
  }

int TPlayground::addPlayer(TPlayer* pla) {
  if(size(pl) >= 100) throw TException("too many players!");
  pla->index = size(pl);
  pla->spx = 1 + (rand() % (sizeX - 2));
  pla->spy = 1 + (rand() % (sizeY - 2));
  pl.push_back(pla);
  return pla->index;
  }

void TPlayground::upTime(int newTime) {
  while(time < newTime) {
    makeBackups();
    int timeObj = NEVER;
    int timeKey = NEVER;
    for(int o=0; o<size(objects); o++) {
      int t = objects[o]->getTime();
      if(t < timeObj) timeObj = t;
      }
    if(kqpos != kqend && kqpos % KQSIZE == kqend % KQSIZE)
      throw TException("I have lost my keys");

    if(kqpos != kqend) timeKey = kqueue[kqpos%KQSIZE].time;
    
    time = newTime;
    
    if(timeKey <= time) time = timeKey;
    
    if(timeObj <= time) time = timeObj;
    
    while(kqpos != kqend && time == kqueue[kqpos%KQSIZE].time) {
      TKeypress& h(kqueue[kqpos%KQSIZE]);
      kqpos++;
      if(h.key == keyJoinGame) {
        bool netplayer = h.player & 0x1000000;
        TPlayer *pla = new TPlayer;
        addPlayer(pla);
        pla->init(netplayer ? h.player & 0xFFFFFF: lps[h.player].color);
        if(!netplayer) {
          lps[h.player].index = pla->index;
          }
        }
      else {
        pl[h.player]->key[int(h.key)] = h.pressed;
        for(int o=0; o<size(objects); o++) if(objects[o] && objects[o]->owner == h.player)
          objects[o]->reactToKey(h.key, h.pressed);
        pl[h.player]->reactToKey(h.key, h.pressed);
        }
      }
    
    for(int o=0; o<size(objects); o++) if(objects[o] && objects[o]->getTime() == time)
      objects[o]->go();

    // delete objects...

    int o0 = 0;
    for(int o=0; o<size(objects); o++) if(objects[o]) {
      if(o0 != o) objects[o0] = objects[o];
      o0++;
      }
    
    if(o0 != size(objects)) objects.resize(o0);
    }
  }

void TPlayground::deleteObject(TPlayObject *obj) {
  for(int t=0; t<size(objects); t++) if(objects[t] == obj) {
    delete obj;
    objects[t] = NULL;
    }
  }

#define MULT 305920673

int TPlayground::rand() { 
  long long s0 = seed & ((1<<30) - 1);
  long long s1 = seed >> 30;
  s0 *= MULT;
  s1 *= MULT;
  seed = s0 + (s1 << 30)+ (s1 >> 30);
  seed += seed >> 60;
  seed &= (1LL << 60) - 1;

  return seed & 0x7FFFFFFF;
  }

