// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// TPlayer and TPlayObject implementation.

TPlayObject *TBuffer::getObject() {
  char type;
  get(type);
  switch(type) { 
    case 1:  return new OBullet(*this);
    case 2:  return new OSnakeHead(*this);
    case 3:  return new OThruster(*this);
    case 4:  return new OTetris(*this);
    case 5:  return new OSnakeTail(*this);
    case 6:  return new OCastle(*this);
    case 7:  return new OCannonball(*this);
    case 8:  return new OPad(*this);
    case 9:  return new OBall(*this);
    case 10: return new OBomber(*this);
    case 11: return new OBomb(*this);
    }
  }

#define numTypes 16

char *typenames[numTypes+1] = {
  "SNAKE",
  "UTRON",
  "UTETRIS",
  "THRUSTER",
  "BOMBER",
  "CASTLE",
  "PAD",
  "BALL",
  "BREAK WALL",
  "PERMA WALL",
  "CLEAR",
  "CANNON",
  "BONUS AMMO",
  "BONUS LEVEL",
  "BONUS OTHER",
  "RANDOM BONUSES",
  "WAIT"
  };

void TPlayObject::killedBy(int player) {
  if(player != owner) g.pl[player]->frags++;
  ow()->typeval = -1;
  g.deleteObject(this);
  }

void TPlayer::init(int c) {
  color = c;
  lives = 0;
  frags = 0;
  score = 0;
  level = 0;
  ammo  = 10;
  typeval=-1;
  typenum= 0;
  for(int k=0; k<8; k++) key[k] = false;
  }

void TPlayer::reactToKey(PlayerKey pk, bool pressed) {
  if(!pressed) return;
  if(typeval == -1) {
    if(pk == keyUp    && spy > 1) spy--;
    if(pk == keyDown  && spy < g.sizeY-2) spy++;
    if(pk == keyLeft  && spx > 1) spx--;
    if(pk == keyRight && spx < g.sizeX-2) spx++;
    if(pk == keyEarth) {
      typenum--; if(typenum==-1) typenum = numTypes-1;
      }
    if(pk == keyAir) {
      typenum++; typenum %= numTypes;
      }
    if(pk == keyWater) {
      typenum = 0;
      }
    if(pk == keyFire) {
      // level is reset for each new addition
      switch(typenum) {
        case 0:
          if(g.map[spy][spx] != cFree) return;
          new OSnakeHead(index, spx, spy, dirAt(spx, spy), true);
          lives++;
          level = 0;
          typeval = 1;
          return;
//new OSnakeHead(2, 50, 30, 3, false);
//new OThruster(1, 30, 50);
//new OTetris(2, 30, 1);

        case 1:
          if(g.map[spy][spx] != cFree) return;
          new OSnakeHead(index, spx, spy, dirAt(spx, spy), false);
          lives++;
          level = 0;
          typeval = 1;
          return;

        case 2:
          new OTetris(index, spx-1, spy);
          lives++;
          level = 0;
          typeval = 1;
          return;

        case 3:
          if(spy < 2 || spy > g.sizeY - 2) return;
          if(spx < 2 || spx > g.sizeX - 2) return;
          new OThruster(index, spx, spy);
          lives++;
          level = 0;
          typeval = 5000;
          return;
        
        case 4:
          new OBomber(index, Ctr(spx), Ctr(spy));
          lives++;
          level = 0;
          ammo = 10;
          typeval = 1;
          return;
        
        case 5:
          if(g.map[spy][spx] != cFree) return;
          g.mapSet(spx, spy, cCastleWall, index);
          new OCastle(index, spx, spy);
          lives++;
          level = 0;
          typeval = 1000;
          return;
          
        case 6:
          if(spy < 4 || spy > g.sizeY - 4) return;
          if(spx < 4 || spx > g.sizeX - 4) return;
          new OPad(index, Ctr(spx), Ctr(spy));
          lives++;
          level = 0;
          typeval = 1;
          return;

        case 7:
          new OBall(index, Ctr(spx), Ctr(spy), g.rand() % 360);
          return;

        case 8:
          if(mapFreeAt(spx, spy)) g.mapSet(spx, spy, cBreakWall, index);
          return;

        case 9:
          if(mapFreeAt(spx, spy)) g.mapSet(spx, spy, cPermaWall, index);
          return;

        case 10:
          if(mapFreeAt(spx, spy)) g.mapSet(spx, spy, cFree, index);
          return;
        
        case 11:
          if(mapFreeBigAt(spx, spy)) placeBig(spx, spy, cCannonF, index);
          return;
          
        case 12:
          if(mapFreeBigAt(spx, spy)) placeBig(spx, spy, cBonusAmmo, index);
          return;
          
        case 13:
          if(mapFreeBigAt(spx, spy)) placeBig(spx, spy, cBonusLevel, index);
          return;
          
        case 14:
          if(mapFreeBigAt(spx, spy)) placeBig(spx, spy, cBonusScore, index);
          return;
          
        case 15:
          placeBigRandomly(cBonusScore, index);
          placeBigRandomly(cBonusLevel, index);
          placeBigRandomly(cBonusAmmo, index);
          return;
        }
      }
    }
  }

char *TPlayer::getTypeName() { return typenames[typenum]; }

void TPlayer::draw() {
  if(typeval == -1) {
    d.setIconColor(color);
    d.spawncursor.drawAtCentered(
      d.ofsX + d.Res * spx + d.Res/2, 
      d.ofsY + d.Res * spy + d.Res/2
      );
    }
  }

