// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// Various simple utility functions.

#include <math.h>

int isSnakeDir(CellType c) {
  if(c == cSnakeUp) return 0;
  if(c == cSnakeRight) return 1;
  if(c == cSnakeDown) return 2;
  if(c == cSnakeLeft) return 3;
  return 4;
  }

CellType snakeDirFor(int d) {
  static CellType c[4] = {cSnakeUp, cSnakeRight, cSnakeDown, cSnakeLeft};
  return c[d];
  }

bool isSnakePart(CellType c) {
  return c == cSnakeHead || c == cSnakeDead ||
    c == cSnakeLeft || c == cSnakeRight || c == cSnakeUp || c == cSnakeDown;
  }

bool isCannon(CellType m) {
  return m == cCannonF || m == cCannonA;
  }

bool isBonus(CellType m) {
  return m == cBonusScore || m == cBonusLevel || m == cBonusAmmo;
  }

bool isBig(CellType m) {
  return isCannon(m) || isBonus(m);
  }

CellType bigType(int x, int y) {
  if(y == 0 || x == 0) return cFree;
  for(int ky=0; ky<2; ky++) for(int kx=0; kx<2; kx++) {
    CellType m = g.map[y-ky][x-kx];
    if(isBig(m))
      return m;
    }
  return cFree;
  }

void clearBig(int x, int y, CellType ct, int owner) {
  for(int ky=0; ky<2; ky++) for(int kx=0; kx<2; kx++) {
    CellType m = g.map[y-ky][x-kx];
    if(isBig(m)) {
      y -= ky;
      x -= kx;
      for(int dy=0; dy<2; dy++) for(int dx=0; dx<2; dx++) {
        g.mapSet(x+dx, y+dy, ct, owner);
        }
      return;
      }
    }
  }

void placeBig(int x, int y, CellType ct, int owner) {
  for(int dy=0; dy<2; dy++) for(int dx=0; dx<2; dx++) {
    g.mapSet(x+dx, y+dy, cBig, owner);
    }
  g.mapSet(x,y, ct, owner);
  }

bool canPlaceBig(int x, int y) {
  return
    g.map[y][x] == cFree && g.map[y+1][x] == cFree &&
    g.map[y][x+1] == cFree && g.map[y+1][x+1] == cFree;
  }

bool placeBigRandomly(CellType ct, int owner) {
  for(int t=0; t<1000; t++) {
    int y = 1 + g.rand() % (g.sizeY - 2);
    int x = 1 + g.rand() % (g.sizeX - 2);
    if(canPlaceBig(x, y)) {
      placeBig(x, y, ct, owner);
      return true;
      }
    }
  return false;
  }

static inline int Ctr(int x) {return (x<<10) + 512;}
static inline int DeCtr(int x) {return x >> 10;}
static inline int CTG(int x) {return ((x*d.Res)>>10);}

TPlayer *TPlayObject::ow() {return g.pl[owner];}

bool before(TKeypress& k1, TKeypress& k2) {
  if(k1.time != k2.time) return k1.time < k2.time;
  if(k1.pressed != k2.pressed) return k1.pressed;
  if(k1.player != k2.player) return k1.player < k2.player;
  return k1.key < k2.key;
  }

int readInt(FILE *f) {
  int i;
  fscanf(f, "%d", &i);
  return i;
  }

string readStr(FILE *f) {
  string res;
  while(1) {
    int k = getc(f);
    if(k < 0) return res;
    else if(k < 32) {if(res != "") return res;}
    else res += char(k);
    }
  }

int degx[360], degy[360];

void initDegs() {
  for(int t=0; t<360; t++) {
    degx[t] = int(sin(t * M_PI / 180) * 128);
    degy[t] = int(-cos(t * M_PI / 180) * 128);
    }
  }

bool mapFreeAt(int x, int y) {
  if(x == 0 || y == 0 || x == g.sizeX-1 || y == g.sizeY-1) return false;
  if(isSnakePart(g.map[y][x])) return false;
  clearBig(x, y, cFree, 0);
  return true;
  }

bool mapFreeBigAt(int x, int y) {
  return mapFreeAt(x,y) && mapFreeAt(x+1, y) && mapFreeAt(x,y+1) && mapFreeAt(x+1,y+1);
  }

int systemSeed() {
  return time(NULL);
  }

// what direction to go from (x,y)
int dirAt(int x, int y) {
  int q = g.sizeY - y;
  int d = 0;
  if(x < q) {d = 1; q = x; }
  if(y < q) {d = 2; q = y; }
  if(g.sizeX-x < q) d = 3;
  return d;
  }

bool isCastlable(CellType m) {
  return m == cFree || m == cBig || isBig(m) || m == cTetris ||
    isSnakePart(m) || m == cExplosion || m == cBomb1 || m == cBomb2;
  }

#define dfDestroyCannons 1

bool destroyAt(int x, int y, int who, int flags=0) {
  switch(g.map[y][x]) { 
    case cFree:
      return false;
    case cSnakeHead:
      g.mapSet(x,y, cSnakeDead, who);
      return true;
    case cSnakeTRON:
    case cTetris:
    case cTetrisDown:
    case cBreakWall:
    case cCastleWall:
      g.mapSet(x,y, cFree, 0);
      return true;
    case cBomb1:
    case cBomb2:
      g.mapSet(x,y, cExplosion, who);
      return true;
    default:
      CellType m = bigType(x,y);
      if(isCannon(m) && ((flags&dfDestroyCannons) || ((g.rand() &7) == 3)))
        clearBig(x, y, cFree, 0);
      return true;
    }
  }

int calcDistance(int ox, int oy) {
  if(ox < 0) ox = -ox;
  if(oy < 0) oy = -oy;
  if(oy > ox) swap(ox, oy);
  return ox+oy/2;
  }
