// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// Main file.

#include "untahris.h"

char *configfile = "config.unt";

#include "tools.cpp"
#include "intface.cpp"
#include "net.cpp"

#undef main

#include "bullet.cpp"
#include "thrust.cpp"
#include "ball.cpp"
#include "snake.cpp"
#include "tetris.cpp"
#include "bomber.cpp"
#include "castle.cpp"

#include "player.cpp"
#include "pground.cpp"

void defaultConfig() {
  lps.clear();
  c.server = "localhost";
  c.sizeX = 50;
  c.sizeY = 30;
  c.res   = 16;
  c.resX  = 800;
  c.resY  = 600;
  c.port  = 5790;
  c.fs    = false;
  c.centered = false;
  
  TLocalPlayer lp;
  lp.color = randColor();
  lp.keys[int(keyLeft )] = 'o';
  lp.keys[int(keyRight)] = 'p';
  lp.keys[int(keyUp   )] = 'q';
  lp.keys[int(keyDown )] = 'a';
  lp.keys[int(keyFire )] = ' ';
  lp.keys[int(keyEarth)] = '1';
  lp.keys[int(keyAir  )] = '2';
  lp.keys[int(keyWater)] = '3';
  lps.push_back(lp);
  
  TLocalPlayer lq;
  lq.color = randColor();
  lq.keys[int(keyLeft )] = SDLK_LEFT;
  lq.keys[int(keyRight)] = SDLK_RIGHT;
  lq.keys[int(keyUp   )] = SDLK_UP;
  lq.keys[int(keyDown )] = SDLK_DOWN;
  lq.keys[int(keyFire )] = SDLK_RETURN;
  lq.keys[int(keyEarth)] = SDLK_INSERT;
  lq.keys[int(keyAir  )] = SDLK_HOME;
  lq.keys[int(keyWater)] = SDLK_PAGEUP;
  lps.push_back(lq);
  }

void writeStr(FILE *f, string s) {
  if(s == "") s = "-";
  fprintf(f, "%s\n", s.c_str());
  }

void writeInt(FILE *f, int i) {
  fprintf(f, "%d\n", i);
  }

void saveConfig() {
  FILE *f = fopen(configfile, "wt");
  if(!f) return;  
  writeStr(f, c.server);
  writeInt(f, int(c.gt));
  writeInt(f, c.port);
  writeInt(f, c.res);
  writeInt(f, c.sizeX);
  writeInt(f, c.sizeY);
  writeInt(f, c.resX);
  writeInt(f, c.resY);
  writeInt(f, int(c.fs));
  writeInt(f, int(c.centered));
  writeInt(f, size(lps));
  for(int p=0; p<size(lps); p++) {
    writeInt(f, lps[p].color);
    for(int k=0; k<8; k++) writeInt(f, lps[p].keys[k]);
    }
  fclose(f);
  }

void loadConfig() {
  FILE *f = fopen(configfile, "rt");
  if(!f) {defaultConfig(); return;}
  c.server = readStr(f);
  c.gt = TGameType(readInt(f));
  c.port = readInt(f);
  c.res = readInt(f);
  c.sizeX = readInt(f);
  c.sizeY = readInt(f);
  c.resX = readInt(f);
  c.resY = readInt(f);
  c.fs = bool(readInt(f));
  c.centered = bool(readInt(f));
  int players = readInt(f);
  lps.clear();
  for(int p=0; p<players; p++) {
    TLocalPlayer newOne;
    newOne.color = readInt(f);
    for(int k=0; k<8; k++) newOne.keys[k] = readInt(f);
    lps.push_back(newOne);
    }
  fclose(f);
  }

int main(int argc, char **argv) {
  g.sizeX = 0;
  g.sizeY = 0;
  d.ofsX = 0;
  d.ofsY = 0;
  d.Res = 16;
  if(argc == 2) configfile = argv[1];
  initDegs();

  try { 
    srand(time(NULL));
    loadConfig();
  
    d.init();
  
    printf("Untahris Common Playground\n");
    printf("copyright (C) 2006 longuens@users.sourceforge.net\n");
    printf("Released under GNU General Public License (see the file \"COPYING\")\n");

    checkConfig();
    d.menuLoop();
  
    saveConfig();
    }
  catch(TException e) {
    printf("Exception catched\n");
    printf("%s/%d/%s\n", e.msg.c_str(), e.data, e.ds.c_str());
    }
  }
